# Copyright 2022 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net
from oslo_log import log as logging
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import event
from ovsdbapp.backend.ovs_idl import idlutils

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils
from neutron.tests.functional import base


LOG = logging.getLogger(__name__)


class WaitForPortGroupDropEvent(event.WaitEvent):
    event_name = 'WaitForPortGroupDropEvent'

    def __init__(self):
        table = 'Port_Group'
        events = (self.ROW_CREATE, self.ROW_UPDATE)
        conditions = (('name', '=', ovn_const.OVN_DROP_PORT_GROUP_NAME),)
        super().__init__(events, table, conditions, timeout=5)


class WaitForChassisCreate(event.WaitEvent):
    event_name = 'WaitForChassisCreate'

    def __init__(self, hostname):
        table = 'Chassis'
        events = (self.ROW_CREATE, )
        conditions = (('hostname', '=', hostname),)
        super().__init__(events, table, conditions, timeout=5)


class TestCreateNeutronPgDrop(base.TestOVNFunctionalBase):
    def _create_and_check_pg_drop(self, wait_event=True):
        pg_event = WaitForPortGroupDropEvent()
        self.mech_driver.nb_ovn.idl.notify_handler.watch_event(pg_event)
        utils.create_neutron_pg_drop()
        if wait_event:
            self.assertTrue(pg_event.wait())
        pg = self.nb_api.pg_get(ovn_const.OVN_DROP_PORT_GROUP_NAME).execute()
        self.assertIsNotNone(pg)
        return pg

    def test_already_existing(self):
        # Make sure pre-fork initialize created the table
        existing_pg = self.nb_api.pg_get(
            ovn_const.OVN_DROP_PORT_GROUP_NAME).execute()
        self.assertIsNotNone(existing_pg)

        # make an attempt to create it again
        pg = self._create_and_check_pg_drop(wait_event=False)
        self.assertEqual(existing_pg.uuid, pg.uuid)

    def test_non_existing(self):
        # Delete the neutron_pg_drop created by pre-fork initialize
        self.nb_api.pg_del(ovn_const.OVN_DROP_PORT_GROUP_NAME).execute()
        pg = self.nb_api.pg_get(ovn_const.OVN_DROP_PORT_GROUP_NAME).execute()
        self.assertIsNone(pg)

        pg = self._create_and_check_pg_drop()

        directions = ['to-lport', 'from-lport']
        matches = ['outport == @neutron_pg_drop && ip',
                   'inport == @neutron_pg_drop && ip']

        # Make sure ACLs are correct
        self.assertEqual(2, len(pg.acls))
        acl1, acl2 = pg.acls

        self.assertEqual('drop', acl1.action)
        self.assertIn(acl1.direction, directions)
        directions.remove(acl1.direction)
        self.assertIn(acl1.match, matches)
        matches.remove(acl1.match)

        self.assertEqual(directions[0], acl2.direction)
        self.assertEqual('drop', acl2.action)
        self.assertEqual(matches[0], acl2.match)


class TestSyncHaChassisGroup(base.TestOVNFunctionalBase):

    def _add_fake_chassis_and_wait(self, hostnames,
                                   enable_chassis_as_gw=None,
                                   enable_chassis_as_extport=None,
                                   other_config=None):
        ch_list = []
        for idx, hostname in enumerate(hostnames):
            # NOTE(ralonsoh): wait until the "Chassis" are created and the
            # event received, to avoid the "HA_Chassis_Group" mangling done in
            # ``ovsdb_monitor.ChassisEvent`` during the next operations.
            ch_event = WaitForChassisCreate(hostname)
            self.sb_api.idl.notify_handler.watch_event(ch_event)
            kwargs = {'azs': []}
            if enable_chassis_as_gw and enable_chassis_as_gw[idx]:
                kwargs['enable_chassis_as_gw'] = True
            if enable_chassis_as_extport and enable_chassis_as_extport[idx]:
                kwargs['enable_chassis_as_extport'] = True
            if other_config and other_config[idx]:
                kwargs['other_config'] = other_config[idx]
            ch_list.append(
                self.add_fake_chassis(hostname, **kwargs))
            ch_event.wait()

        return ch_list

    def test_sync_ha_chassis_group_network(self):
        net = self._make_network(self.fmt, 'n1', True)['network']
        port_id = 'fake-port-id'
        hcg_name = utils.ovn_name(net['id'])
        ch_list = self._add_fake_chassis_and_wait(
            ('host1', 'host2', 'host3'),
            enable_chassis_as_gw=[True, True, False])

        with self.nb_api.transaction(check_error=True) as txn:
            utils.sync_ha_chassis_group_network(
                self.context, self.nb_api, self.sb_api,
                port_id, net['id'], txn)

        ha_chassis = self.nb_api.db_find('HA_Chassis').execute(
            check_error=True)
        ha_chassis_names = [hc['chassis_name'] for hc in ha_chassis]
        self.assertEqual(2, len(ha_chassis))
        self.assertEqual(sorted([ch_list[0], ch_list[1]]),
                         sorted(ha_chassis_names))

        hcg = self.nb_api.ha_chassis_group_get(hcg_name).execute(
            check_error=True)
        self.assertEqual(hcg_name, hcg.name)
        ha_chassis_exp = sorted([str(hc['_uuid']) for hc in ha_chassis])
        ha_chassis_ret = sorted([str(hc.uuid) for hc in hcg.ha_chassis])
        self.assertEqual(ha_chassis_exp, ha_chassis_ret)

        # Delete one GW chassis and resync the HA chassis group associated to
        # the same network. The method will now not create again the existing
        # HA Chassis Group register but will update the "ha_chassis" list.
        self.del_fake_chassis(ch_list[1])
        with self.nb_api.transaction(check_error=True) as txn:
            utils.sync_ha_chassis_group_network(
                self.context, self.nb_api, self.sb_api, port_id,
                net['id'], txn)

        ha_chassis = self.nb_api.db_find('HA_Chassis').execute(
            check_error=True)
        ha_chassis_names = [hc['chassis_name'] for hc in ha_chassis]
        self.assertEqual(1, len(ha_chassis))
        self.assertEqual([ch_list[0]], ha_chassis_names)

        hcg = self.nb_api.ha_chassis_group_get(hcg_name).execute(
            check_error=True)
        self.assertEqual(hcg_name, hcg.name)
        ha_chassis_exp = str(ha_chassis[0]['_uuid'])
        ha_chassis_ret = str(hcg.ha_chassis[0].uuid)
        self.assertEqual(ha_chassis_exp, ha_chassis_ret)

    def test_sync_ha_chassis_group_network_extport(self):
        # Create a network and an external port
        net = self._make_network(self.fmt, 'n1', True)['network']
        port_data = {
            'port': {'network_id': net['id'],
                     'project_id': self._project_id,
                     portbindings.VNIC_TYPE: portbindings.VNIC_DIRECT}}
        port_req = self.new_create_request('ports', port_data, self.fmt)
        port_res = port_req.get_response(self.api)
        port = self.deserialize(self.fmt, port_res)['port']

        # Add 3 chassis, two eligible for hosting the external port
        ch_list = self._add_fake_chassis_and_wait(
            ('host1', 'host2', 'host3'),
            enable_chassis_as_extport=[True, True, False])

        # Invoke the sync method
        with self.nb_api.transaction(check_error=True) as txn:
            hcg, _ = utils.sync_ha_chassis_group_network(
                self.context, self.nb_api, self.sb_api, port['id'],
                net['id'], txn)
            # It is needed to assign the HCG to the LSP. When the port is
            # deleted, the external port HCG associated will be deleted too.
            txn.add(
                self.nb_api.set_lswitch_port(port['id'], ha_chassis_group=hcg))

        # Assert only the eligible chassis are present in HA Chassis
        ha_chassis = self.nb_api.db_find('HA_Chassis').execute(
            check_error=True)
        ha_chassis_names = [hc['chassis_name'] for hc in ha_chassis]
        self.assertEqual(2, len(ha_chassis))
        self.assertEqual(sorted([ch_list[0], ch_list[1]]),
                         sorted(ha_chassis_names))

        # Assert the HA Chassis Group has the correct name and the
        # eligible chassis are included in it
        hcg_name = utils.ovn_extport_chassis_group_name(port['id'])
        hcg = self.nb_api.ha_chassis_group_get(hcg_name).execute(
            check_error=True)
        self.assertEqual(hcg_name, hcg.name)
        ha_chassis_exp = sorted([str(hc['_uuid']) for hc in ha_chassis])
        ha_chassis_ret = sorted([str(hc.uuid) for hc in hcg.ha_chassis])
        self.assertEqual(ha_chassis_exp, ha_chassis_ret)

        # Delete one eligible Chassis and resync the HA chassis group
        # associated to the external port. The method should not re-create
        # the existing HA Chassis Group but only update the "ha_chassis" list
        self.del_fake_chassis(ch_list[1])
        with self.nb_api.transaction(check_error=True) as txn:
            utils.sync_ha_chassis_group_network(
                self.context, self.nb_api, self.sb_api, port['id'],
                net['id'], txn)

        # Assert the chassis deletion reflects in the HA Chassis and
        # HA Chassis Group
        ha_chassis = self.nb_api.db_find('HA_Chassis').execute(
            check_error=True)
        ha_chassis_names = [hc['chassis_name'] for hc in ha_chassis]
        self.assertEqual(1, len(ha_chassis))
        self.assertEqual([ch_list[0]], ha_chassis_names)

        hcg = self.nb_api.ha_chassis_group_get(hcg_name).execute(
            check_error=True)
        self.assertEqual(hcg_name, hcg.name)
        ha_chassis_exp = str(ha_chassis[0]['_uuid'])
        ha_chassis_ret = str(hcg.ha_chassis[0].uuid)
        self.assertEqual(ha_chassis_exp, ha_chassis_ret)

        # Delete the external port, assert that the HA Chassis and HA Chassis
        # Group were also deleted
        self.plugin.delete_port(self.context, port['id'])
        ha_chassis = self.nb_api.db_find('HA_Chassis').execute(
            check_error=True)
        self.assertEqual(0, len(ha_chassis))
        self.assertRaises(
            idlutils.RowNotFound,
            self.nb_api.ha_chassis_group_get(hcg_name).execute,
            check_error=True)

    def _test_sync_unify_ha_chassis_group_network(self, create_hcg=False):
        def print_error(hcg):
            LOG.error('HA_Chassis in HCG %s', hcg.name)
            for hc in hcg.ha_chassis:
                LOG.error('  - Chassis name: %s, priority: %s',
                          hc.chassis_name, hc.priority)

        physnet = 'physnet1'
        net_ext_args = {provider_net.NETWORK_TYPE: 'vlan',
                        provider_net.PHYSICAL_NETWORK: physnet,
                        external_net.EXTERNAL: True}
        net_ext = self._make_network(self.fmt, 'test-ext-net', True,
                                     as_admin=True,
                                     arg_list=tuple(net_ext_args.keys()),
                                     **net_ext_args)['network']
        other_config = {'ovn-bridge-mappings': physnet + ':br-ex'}
        ch_list = self._add_fake_chassis_and_wait(
            ('host1', 'host2', 'host3'),
            enable_chassis_as_gw=[True, True, True],
            other_config=[other_config, other_config, None])

        # Create a pre-existing HCG.
        group_name = utils.ovn_name(net_ext['id'])
        if create_hcg:
            chassis_list = [self.sb_api.lookup('Chassis', ch_list[1])]
            hcg_info = utils.HAChassisGroupInfo(
                group_name=group_name, chassis_list=chassis_list,
                az_hints=[], ignore_chassis=set(), external_ids={})
            with self.nb_api.transaction(check_error=True) as txn:
                utils._sync_ha_chassis_group(self.nb_api, hcg_info, txn)
            hcg = self.nb_api.lookup('HA_Chassis_Group', group_name)
            try:
                self.assertEqual(1, len(hcg.ha_chassis))
                self.assertEqual(ovn_const.HA_CHASSIS_GROUP_HIGHEST_PRIORITY,
                                 hcg.ha_chassis[0].priority)
            except AssertionError as exc:
                print_error(hcg)
                raise exc

        # Invoke the sync method
        chassis_prio = {ch_list[0]: 10, ch_list[1]: 20, ch_list[2]: 30}
        with self.nb_api.transaction(check_error=True) as txn:
            utils.sync_ha_chassis_group_network_unified(
                self.context, self.nb_api, self.sb_api, net_ext['id'],
                'router-id', chassis_prio, txn)

        hcg = self.nb_api.lookup('HA_Chassis_Group', group_name)
        try:
            self.assertEqual(3, len(hcg.ha_chassis))
            for hc in hcg.ha_chassis:
                self.assertEqual(chassis_prio[hc.chassis_name], hc.priority)
        except AssertionError as exc:
            print_error(hcg)
            raise exc

    def test_sync_unify_ha_chassis_group_network_no_hcg(self):
        self._test_sync_unify_ha_chassis_group_network()

    def test_sync_unify_ha_chassis_group_network_existing_hcg(self):
        self._test_sync_unify_ha_chassis_group_network(create_hcg=True)


@utils.ovn_context()
def method_with_idl_and_default_txn(ls_name, idl, txn=None):
    txn.add(idl.ls_add(ls_name))


@utils.ovn_context()
def method_with_txn_and_default_idl(ls_name, txn, idl=None):
    # NOTE(ralonsoh): the test with the default "idl" cannot be executed. A
    # default value should be provided in a non-testing implementation.
    txn.add(idl.ls_add(ls_name))


@utils.ovn_context()
def method_with_idl_and_txn(ls_name, idl, txn):
    txn.add(idl.ls_add(ls_name))


@utils.ovn_context(txn_var_name='custom_txn', idl_var_name='custom_idl')
def method_with_custom_idl_and_custom_txn(ls_name, custom_idl, custom_txn):
    custom_txn.add(custom_idl.ls_add(ls_name))


@utils.ovn_context()
def update_ls(ls_name, idl, txn):
    txn.add(idl.db_set('Logical_Switch', ls_name,
                       ('external_ids', {'random_key': 'random_value'})
                       )
            )


@ddt.ddt()
class TestOvnContext(base.TestOVNFunctionalBase):

    scenarios = (
        {'name': 'idl_and_default_txn',
         'method': method_with_idl_and_default_txn,
         '_args': ['ls_name', 'idl'], '_kwargs': ['txn']},
        {'name': 'idl_and_default_txn__positional_txn',
         'method': method_with_idl_and_default_txn,
         '_args': ['ls_name', 'idl', 'txn'], '_kwargs': []},
        {'name': 'idl_and_default_txn__default_txn',
         'method': method_with_idl_and_default_txn,
         '_args': ['ls_name', 'idl'], '_kwargs': []},

        {'name': 'txn_and_default_idl',
         'method': method_with_txn_and_default_idl,
         '_args': ['ls_name', 'txn'], '_kwargs': ['idl']},
        {'name': 'txn_and_default_idl__positional_idl',
         'method': method_with_txn_and_default_idl,
         '_args': ['ls_name', 'txn', 'idl'], '_kwargs': []},

        {'name': 'txn_and_idl',
         'method': method_with_idl_and_txn,
         '_args': ['ls_name', 'idl', 'txn'], '_kwargs': []},

        {'name': 'custom_idl_and_custom_txn',
         'method': method_with_custom_idl_and_custom_txn,
         '_args': ['ls_name', 'custom_idl', 'custom_txn'], '_kwargs': []},
    )

    scenarios2 = (
        {'name': method_with_idl_and_default_txn.__name__,
         'method': method_with_idl_and_default_txn},
        {'name': method_with_txn_and_default_idl.__name__,
         'method': method_with_txn_and_default_idl},
        {'name': method_with_idl_and_txn.__name__,
         'method': method_with_idl_and_txn},
        {'name': method_with_custom_idl_and_custom_txn.__name__,
         'method': method_with_custom_idl_and_custom_txn},
    )

    @ddt.unpack
    @ddt.named_data(*scenarios)
    def test_with_transaction(self, method, _args, _kwargs):
        ls_name = uuidutils.generate_uuid()
        custom_idl = idl = self.nb_api
        with self.nb_api.transaction(check_error=True) as txn:
            custom_txn = txn
            _locals = locals()
            args = [_locals[_arg] for _arg in _args]
            kwargs = {_kwarg: _locals[_kwarg] for _kwarg in _kwargs}
            # Create a LS and update it.
            method(*args, **kwargs)
            update_ls(ls_name, self.nb_api, txn)

        ls = self.nb_api.lookup('Logical_Switch', ls_name)
        self.assertEqual('random_value', ls.external_ids['random_key'])

    @ddt.unpack
    @ddt.named_data(*scenarios)
    def test_without_transaction(self, method, _args, _kwargs):
        ls_name = uuidutils.generate_uuid()
        custom_idl = idl = self.nb_api
        custom_txn = txn = None
        _locals = locals()
        args = [_locals[_arg] for _arg in _args]
        kwargs = {_kwarg: _locals[_kwarg] for _kwarg in _kwargs}
        # Create a LS and update it.
        method(*args, **kwargs)
        update_ls(ls_name, self.nb_api, txn)

        ls = self.nb_api.lookup('Logical_Switch', ls_name)
        self.assertEqual('random_value', ls.external_ids['random_key'])

    @ddt.unpack
    @ddt.named_data(*scenarios2)
    def test_needed_parameters(self, method):
        self.assertRaises(RuntimeError, method, uuidutils.generate_uuid(),
                          None, None)


class TestGetLogicalRouterPortHAChassis(base.TestOVNFunctionalBase):
    def _create_network_and_port(self):
        kwargs = {external_net.EXTERNAL: True, 'as_admin': True}
        net = self._make_network(self.fmt, 'n1', True, **kwargs)['network']
        port_data = {'port': {'network_id': net['id'],
                              'project_id': self._project_id,}}
        port_req = self.new_create_request('ports', port_data, self.fmt)
        port_res = port_req.get_response(self.api)
        return self.deserialize(self.fmt, port_res)['port']

    def _create_gw_chassis(self, num_chassis):
        chassis = []
        for _ in range(num_chassis):
            chassis.append(self.add_fake_chassis(
                uuidutils.generate_uuid(), azs=[],
                enable_chassis_as_gw=True))
        return chassis

    def _create_router(self, network_id):
        gw_info = {'network_id': network_id}
        router = {'router': {'name': uuidutils.generate_uuid(),
                             'admin_state_up': True,
                             'project_id': self._project_id,
                             'external_gateway_info': gw_info}}
        return self.l3_plugin.create_router(self.context, router)

    def _set_lrp_hcg(self, gw_port_id, hcg):
        lrp_name = utils.ovn_lrouter_port_name(gw_port_id)
        self.nb_api.db_set(
            'Logical_Router_Port', lrp_name,
            ('ha_chassis_group', hcg.uuid)).execute()
        return self.nb_api.lookup('Logical_Router_Port', lrp_name)

    def _get_router_hcg(self, router_id):
        hcg_name = utils.ovn_name(router_id)
        return self.nb_api.lookup('HA_Chassis_Group', hcg_name)

    def _check_chassis(self, ha_chassis, expected_chassis, priorities=None):
        length = len(priorities) if priorities else len(expected_chassis)
        self.assertEqual(length, len(ha_chassis))
        ch_priorities = set([])
        for hc in ha_chassis:
            self.assertIn(hc[0], expected_chassis)
            ch_priorities.add(hc[1])
        self.assertEqual(length, len(ch_priorities))
        if priorities:
            for ch_priority in ch_priorities:
                self.assertIn(ch_priority, priorities)

    def test_get_ha_chassis(self):
        port = self._create_network_and_port()
        ch_list = self._create_gw_chassis(5)
        router = self._create_router(port['network_id'])
        hcg = self._get_router_hcg(router['id'])
        lrp = self._set_lrp_hcg(router['gw_port_id'], hcg)

        ha_chassis = utils.get_logical_router_port_ha_chassis(self.nb_api, lrp)
        self._check_chassis(ha_chassis, ch_list)

    def test_get_ha_chassis_priorities(self):
        port = self._create_network_and_port()
        ch_list = self._create_gw_chassis(5)
        router = self._create_router(port['network_id'])
        hcg = self._get_router_hcg(router['id'])
        lrp = self._set_lrp_hcg(router['gw_port_id'], hcg)

        prio = [ovn_const.HA_CHASSIS_GROUP_HIGHEST_PRIORITY,
                ovn_const.HA_CHASSIS_GROUP_HIGHEST_PRIORITY - 1]
        ha_chassis = utils.get_logical_router_port_ha_chassis(
            self.nb_api, lrp, priorities=prio)
        self._check_chassis(ha_chassis, ch_list, priorities=prio)



# Copyright (c) 2024 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import quota_check_limit
from neutron_lib.api import extensions


# NOTE(ralonsoh): once [1] is merged, use
# ``neutron_lib.api.definitions.quota_check_limit_default`` instead.
# [1] https://review.opendev.org/c/openstack/neutron-lib/+/926777
ALIAS = 'quota-check-limit-default'
IS_SHIM_EXTENSION = True
IS_STANDARD_ATTR_EXTENSION = False
NAME = 'Quota engine limit check by default'
DESCRIPTION = ('By default, the Neutron quota engine checks the resource '
               'usage before applying a new quota limit')
UPDATED_TIMESTAMP = '2024-08-21T16:00:00-00:00'
RESOURCE_ATTRIBUTE_MAP = {}
SUB_RESOURCE_ATTRIBUTE_MAP = {}
ACTION_MAP = {}
REQUIRED_EXTENSIONS = [quota_check_limit.ALIAS]
OPTIONAL_EXTENSIONS = []
ACTION_STATUS = {}


class Quota_check_limit_default(extensions.APIExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return NAME

    @classmethod
    def get_alias(cls):
        return ALIAS

    @classmethod
    def get_description(cls):
        return DESCRIPTION

    @classmethod
    def get_updated(cls):
        return UPDATED_TIMESTAMP

    def get_required_extensions(self):
        return REQUIRED_EXTENSIONS

    def get_optional_extensions(self):
        return []

    @classmethod
    def get_extended_resources(cls, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        return {}



# Copyright 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

from oslo_config import fixture as fixture_config
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import event
from ovsdbapp.backend.ovs_idl import idlutils

from neutron.agent.ovn.agent import ovn_neutron_agent
from neutron.agent.ovn.agent import ovsdb as agent_ovsdb
from neutron.agent.ovn.metadata import agent as metadata_agent
from neutron.agent.ovn.metadata import server_socket
from neutron.agent.ovsdb import impl_idl
from neutron.common.ovn import constants as ovn_const
from neutron.common import utils as n_utils
from neutron.tests.common import net_helpers
from neutron.tests.functional import base


TEST_EXTENSION = 'testing'
METADATA_EXTENSION = 'metadata'
EXTENSION_NAMES = {TEST_EXTENSION: 'Fake OVN agent extension',
                   METADATA_EXTENSION: 'Metadata OVN agent extension',
                   }


class ChassisPrivateUpdateEvent(event.WaitEvent):
    def __init__(self, chassis_private, timeout=5):
        table = 'Chassis_Private'
        events = (self.ROW_UPDATE,)
        conditions = (('name', '=', chassis_private),)
        super().__init__(events, table, conditions, timeout=timeout)


class TestOVNNeutronAgentBase(base.TestOVNFunctionalBase):

    def setUp(self, extensions, **kwargs):
        super().setUp(**kwargs)
        self.host_name = 'host-' + uuidutils.generate_uuid()[:5]
        self.chassis_name = self.add_fake_chassis(self.host_name)
        self.extensions = extensions
        self.mock_chassis_name = mock.patch.object(
            agent_ovsdb, 'get_own_chassis_name',
            return_value=self.chassis_name).start()
        self.ovs_idl_events = []
        with mock.patch.object(metadata_agent.MetadataAgent,
                               '_get_own_chassis_name',
                               return_value=self.chassis_name):
            self.ovn_agent = self._start_ovn_neutron_agent()

    def _check_loaded_and_started_extensions(self, ovn_agent):
        for _ext in self.extensions:
            loaded_ext = ovn_agent[_ext]
            self.assertEqual(EXTENSION_NAMES.get(_ext), loaded_ext.name)
            self.assertTrue(loaded_ext.is_started)

    def _create_ovs_idl(self, ovn_agent):
        for extension in ovn_agent.ext_manager:
            self.ovs_idl_events += extension.obj.ovs_idl_events
        self.ovs_idl_events = [e(ovn_agent) for e in
                                         set(self.ovs_idl_events)]
        ovsdb = impl_idl.api_factory()
        ovsdb.idl.notify_handler.watch_events(self.ovs_idl_events)

        ovn_agent.ext_manager_api.ovs_idl = ovsdb
        return ovsdb

    def _clear_events_ovs_idl(self):
        self.ovn_agent.ovs_idl.idl_monitor.notify_handler.unwatch_events(
            self.ovs_idl_events)

    def _start_ovn_neutron_agent(self):
        conf = self.useFixture(fixture_config.Config()).conf
        conf.set_override('extensions', ','.join(self.extensions),
                          group='agent')
        ovn_nb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('nb')
        conf.set_override('ovn_nb_connection', [ovn_nb_db], group='ovn')
        ovn_sb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('sb')
        conf.set_override('ovn_sb_connection', [ovn_sb_db], group='ovn')

        agt = ovn_neutron_agent.OVNNeutronAgent(conf)
        agt.test_ovs_idl = []
        agt.test_ovn_sb_idl = []
        agt.test_ovn_nb_idl = []
        # NOTE(ralonsoh): it is needed to ``UnixDomainMetadataProxy.wait``
        # method in eventlet environments in order not to block the execution.
        # Once eventlet is completely removed, this mock can be deleted.
        with mock.patch.object(ovn_neutron_agent.OVNNeutronAgent, 'wait'), \
                mock.patch.object(server_socket.UnixDomainMetadataProxy,
                                  'wait'), \
                mock.patch.object(ovn_neutron_agent.OVNNeutronAgent,
                                  '_load_ovs_idl') as mock_load_ovs_idl:
            agt._initialize_ext_manager()
            mock_load_ovs_idl.return_value = self._create_ovs_idl(agt)
            agt.start()
            external_ids = agt.sb_idl.db_get(
                'Chassis_Private', agt.chassis, 'external_ids').execute(
                check_error=True)
            self.assertEqual(
                external_ids[ovn_const.OVN_AGENT_NEUTRON_SB_CFG_KEY],
                '0')

        self._check_loaded_and_started_extensions(agt)
        self.addCleanup(self._clear_events_ovs_idl)
        if agt.ext_manager_api.sb_idl:
            self.addCleanup(agt.ext_manager_api.sb_idl.ovsdb_connection.stop)
        if agt.ext_manager_api.nb_idl:
            self.addCleanup(agt.ext_manager_api.nb_idl.ovsdb_connection.stop)
        return agt


class TestOVNNeutronAgent(TestOVNNeutronAgentBase):
    def setUp(self, **kwargs):
        super().setUp(extensions=[METADATA_EXTENSION], **kwargs)

    def test_chassis_private_create_event(self):
        def _check_chassis_private():
            try:
                ext_ids = self.ovn_agent.sb_idl.db_get(
                    'Chassis_Private', self.chassis_name,
                    'external_ids').execute(check_error=True)
                neutron_id = (ext_ids.get(ovn_const.OVN_AGENT_NEUTRON_ID_KEY)
                              is not None)
                agt_cfg = (ext_ids.get(ovn_const.OVN_AGENT_NEUTRON_SB_CFG_KEY)
                           is not None)
                bridge = (ext_ids.get(ovn_const.OVN_AGENT_OVN_BRIDGE)
                          is not None)
                return neutron_id and agt_cfg and bridge
            except idlutils.RowNotFound:
                return False

        # If the "Chassis_Private" register is deleted and created again,
        # the agent should be able to re-register itself.
        self.ovn_agent.sb_idl.chassis_del(self.chassis_name).execute(
            check_error=True)
        self.add_fake_chassis(self.host_name, name=self.chassis_name)

        n_utils.wait_until_true(_check_chassis_private, timeout=10)


class TestOVNNeutronAgentFakeAgent(TestOVNNeutronAgentBase):

    def setUp(self, **kwargs):
        super().setUp(extensions=[TEST_EXTENSION], **kwargs)

    def test_ovs_and_ovs_events(self):
        # Test the OVS IDL is attending the provided events.
        bridge = self.useFixture(net_helpers.OVSBridgeFixture()).bridge
        exc = Exception('Bridge %s not added or not detected by '
                        'OVSInterfaceEvent' % bridge)
        n_utils.wait_until_true(
            lambda: bridge.br_name in self.ovn_agent.test_ovs_idl,
            timeout=10, exception=exc)

        # Test the OVN SB IDL is attending the provided events. The chassis is
        # created before the OVN SB IDL connection is created but the creation
        # event is received during the subscription.
        exc = Exception('Chassis %s not added or not detected by '
                        'OVNSBChassisEvent' % self.chassis_name)
        n_utils.wait_until_true(
            lambda: self.chassis_name in self.ovn_agent.test_ovn_sb_idl,
            timeout=10, exception=exc)

        # Test the OVN SN IDL is attending the provided events.
        lswitch_name = 'ovn-' + uuidutils.generate_uuid()
        self.nb_api.ls_add(lswitch_name).execute(check_error=True)
        exc = Exception('Logical Switch %s not added or not detected by ')
        n_utils.wait_until_true(
            lambda: lswitch_name in self.ovn_agent.test_ovn_nb_idl,
            timeout=10, exception=exc)


class TestOVNNeutronAgentMetadataExtension(TestOVNNeutronAgentBase):

    def setUp(self, **kwargs):
        super().setUp(extensions=[METADATA_EXTENSION], **kwargs)

    def test_check_metadata_started(self):
        def check_extids(expected_ext_ids, chassis_name):
            ch_private = self.sb_api.lookup('Chassis_Private', chassis_name)
            return expected_ext_ids == ch_private.external_ids

        # Check the metadata extension is registered.
        chassis_id = uuid.UUID(self.chassis_name)
        ovn_agent_id = uuid.uuid5(chassis_id, 'ovn_agent')
        ext_ids = {ovn_const.OVN_AGENT_OVN_BRIDGE: 'br-int',
                   ovn_const.OVN_AGENT_NEUTRON_ID_KEY: str(ovn_agent_id),
                   ovn_const.OVN_AGENT_NEUTRON_SB_CFG_KEY: '0',
                   }
        n_utils.wait_until_true(
            lambda: check_extids(ext_ids, self.chassis_name),
            timeout=10)

        # Check Unix proxy is running.
        metadata_extension = self.ovn_agent[METADATA_EXTENSION]
        self.assertIsNotNone(metadata_extension._proxy._server)

    def test__cleanup_previous_tags(self):
        external_ids = {
            ovn_const.OVN_AGENT_METADATA_SB_CFG_KEY: '1',
            ovn_const.OVN_AGENT_METADATA_DESC_KEY: 'description',
            ovn_const.OVN_AGENT_METADATA_ID_KEY: uuidutils.generate_uuid()}
        self.ovn_agent.sb_idl.db_set(
            'Chassis_Private', self.ovn_agent.chassis,
            ('external_ids', external_ids)).execute(check_error=True)

        cp_event = ChassisPrivateUpdateEvent(self.chassis_name)
        self.ovn_agent.sb_idl.idl.notify_handler.watch_event(cp_event)
        self.ovn_agent._cleanup_previous_tags()
        self.assertTrue(cp_event.wait())

        external_ids = self.ovn_agent.sb_idl.db_get(
            'Chassis_Private', self.ovn_agent.chassis,
            'external_ids').execute(check_error=True)
        for _key in (ovn_const.OVN_AGENT_METADATA_SB_CFG_KEY,
                     ovn_const.OVN_AGENT_METADATA_DESC_KEY,
                     ovn_const.OVN_AGENT_METADATA_ID_KEY):
            self.assertNotIn(_key, external_ids)

        # Just in case, check that we are NOT deleting the needed tags.
        # NOTE(ralonsoh): OVN_AGENT_METADATA_ID_KEY is missing here, there is
        # a bug to add it (LP#2118876)
        self.assertIn(ovn_const.OVN_AGENT_NEUTRON_SB_CFG_KEY, external_ids)



# Copyright 2025 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.agent.ovn.extensions import extension_manager
from neutron.tests import base


class OVNExtensionEventNoExtensionName(extension_manager.OVNExtensionEvent):
    pass


class TestOVNExtensionEvent(base.BaseTestCase):
    def test_class_with_no_extension_name(self):
        try:
            OVNExtensionEventNoExtensionName()
        except Exception as exc:
            msg = ('The class OVNExtensionEventNoExtensionName has no '
                   'extension name defined.')
            self.assertEqual(msg, str(exc))



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import os
import tempfile

from oslo_config import cfg
from oslo_utils import timeutils
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp import event
from ovsdbapp import venv

from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.conf.services import bgp as bgp_config
from neutron.services.bgp import ovn as bgp_ovn
from neutron.tests.functional import base as n_base
from neutron.tests.functional.services.bgp import fixtures


class OvsTestIdl(connection.OvsdbIdl):
    tables = ['Open_vSwitch', 'Bridge', 'Port', 'Interface']

    def __init__(self, connection_string):
        helper = idlutils.get_schema_helper(connection_string, 'Open_vSwitch')
        for table in self.tables:
            helper.register_table(table)
        self.notify_handler = event.RowEventHandler()
        super().__init__(connection_string, helper)

    def notify(self, event, row, updates=None):
        self.notify_handler.notify(event, row, updates)


def requires_ovn_version_with_bgp():
    def outer(f):
        @functools.wraps(f)
        def inner(self, *args, **kwargs):
            if not self._is_bgp_supported():
                raise self.skipException(
                    "Used OVN version does not have BGP support")
            return f(self, *args, **kwargs)
        return inner
    return outer


def is_policy_output_port_column_supported(idl):
    # OVN added the output_port column to the Logical_Router_Policy table in
    # v26.03 commit 75681638854291c412a46dbea81bd7e6a6128a3b
    return idlutils.table_has_column(
        idl, 'Logical_Router_Policy', 'output_port')


class BaseBgpIDLTestCase(n_base.BaseLoggingTestCase):
    schemas = []
    idl_schema_map = {
        'OVN_Northbound': bgp_ovn.OvnNbIdl,
        'OVN_Southbound': bgp_ovn.OvnSbIdl,
        'Open_vSwitch': OvsTestIdl,
    }

    def setUp(self):
        ovn_conf.register_opts()
        bgp_config.register_opts(cfg.CONF)
        super().setUp()
        self.setup_venv()
        self.create_idls()

    def create_connection(self, schema):
        idl = self.idl_schema_map[schema](self._schema_map[schema])
        return connection.Connection(idl, timeout=10)

    def setup_venv(self):
        ovsvenv = venv.OvsOvnVenvFixture(
            tempfile.mkdtemp(),
            ovsdir=os.getenv('OVS_SRCDIR'),
            ovndir=os.getenv('OVN_SRCDIR'),
            remove=True)

        self.useFixture(ovsvenv)

        self._schema_map = {
            'OVN_Northbound': ovsvenv.ovnnb_connection,
            'OVN_Southbound': ovsvenv.ovnsb_connection,
            'Open_vSwitch': ovsvenv.ovs_connection,
        }

    def _is_bgp_supported(self):
        # Look at the Southbound IDL to check if Advertised_Route table exists
        helper = idlutils.get_schema_helper(
            self._schema_map['OVN_Southbound'], 'OVN_Southbound')
        return 'Advertised_Route' in helper.schema_json['tables']

    def create_idls(self):
        for schema in self.schemas:
            connection = self.create_connection(schema)
            if schema == 'OVN_Northbound':
                self.nb_api = self.useFixture(
                    fixtures.OvnNbIdlApiFixture(connection)).obj
            elif schema == 'OVN_Southbound':
                self.sb_api = self.useFixture(
                    fixtures.OvnSbIdlApiFixture(connection)).obj
            elif schema == 'Open_vSwitch':
                self.ovs_api = self.useFixture(
                    fixtures.OvsApiFixture(connection)).obj


class BaseBgpNbIdlTestCase(BaseBgpIDLTestCase):
    schemas = ['OVN_Northbound']


class BaseBgpSbIdlTestCase(BaseBgpIDLTestCase):
    schemas = ['OVN_Southbound']

    def setUp(self):
        bgp_ovn.OvnSbIdl.tables = ('Chassis', 'Encap', 'Chassis_Private')
        try:
            super().setUp()
        finally:
            bgp_ovn.OvnSbIdl.tables = bgp_ovn.OVN_SB_TABLES

    def add_fake_chassis(self, name, ip, external_ids=None):
        external_ids = external_ids or {}
        chassis = self.sb_api.chassis_add(
            name, ['geneve'], ip).execute(check_error=True)

        nb_cfg_timestamp = timeutils.utcnow_ts() * 1000
        self.sb_api.db_create(
            'Chassis_Private', name=name,
            chassis=chassis.uuid, nb_cfg_timestamp=nb_cfg_timestamp,
            external_ids=external_ids
        ).execute(check_error=True)

        return self.sb_api.db_list_rows(
            'Chassis_Private', [name]).execute(check_error=True)[0]


class BaseBgpTestCase(BaseBgpSbIdlTestCase):
    schemas = ['OVN_Northbound', 'OVN_Southbound']



# Copyright (c) 2015 UnitedStack, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import signal

import jinja2

from neutron_lib.utils import file as file_utils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils.fileutils import ensure_tree

from neutron._i18n import _
from neutron.agent.linux import external_process
from neutron.agent.linux import utils
from neutron.common import utils as common_utils

CONNTRACKD_SERVICE_NAME = 'conntrackd'
SIGTERM_TIMEOUT = 5

# Unix socket path length is limited to 107 characters in the
# conntrackd source code. See UNIX_PATH_MAX constant in include/local.h
UNIX_PATH_MAX = 107

LOG = logging.getLogger(__name__)


CONFIG_TEMPLATE = jinja2.Template(
    """
General {
    HashSize {{ hash_size }}
    HashLimit {{ hash_limit }}
    Syslog on
    LockFile {{ lockfile_path }}
    UNIX {
        Path {{ socket_path }}
        Backlog {{ unix_backlog }}
    }
    SocketBufferSize {{ socket_buffer_size }}
    SocketBufferSizeMaxGrown {{ socket_buffer_size_max_grown }}
    Filter From Kernelspace {
        Protocol Accept {
{%- for proto in protocol_accept %}
            {{ proto }}
{%- endfor %}
        }
        Address Ignore {
{%- for version, addr in address_ignore %}
            IPv{{ version }}_address {{ addr }}
{%- endfor %}
        }
    }
}
Sync {
    Mode FTFW {
    }
    Multicast Default {
        IPv4_address {{ ipv4_mcast_addr }}
        IPv4_interface {{ ipv4_interface }}
        Group {{ mcast_group }}
        Interface {{ interface }}
        SndSocketBuffer {{ snd_socket_buffer }}
        RcvSocketBuffer {{ rcv_socket_buffer }}
        Checksum on
    }
}
""")


class ConntrackdManager:
    """Wrapper for conntrackd.

    This wrapper permits to write conntrackd config file,
    to start/restart conntrackd process.

    """

    def __init__(self, resource_id, process_monitor, agent_conf,
                 mcast_iface_addr, ha_vr_id, ha_iface, namespace=None):
        self.resource_id = resource_id
        self.process_monitor = process_monitor
        self.agent_conf = agent_conf
        self.mcast_iface_addr = mcast_iface_addr
        self.ha_vr_id = ha_vr_id
        self.ha_iface = ha_iface
        self.namespace = namespace

    def get_full_config_file_path(self, filename, maxlen=255):
        # Maximum PATH length for most paths in conntrackd is limited to
        # 255 characters.
        conf_dir = self.get_conf_dir()
        ensure_tree(conf_dir, 0o755)
        path = os.path.join(conf_dir, filename)

        if len(path) > maxlen:
            raise ValueError(_('Configuration file path "%(path)s" exceeds '
                               'maximum length of %(maxlen)i characters.') %
                             {'path': path,
                              'maxlen': maxlen})

        return path

    def get_conf_dir(self):
        confs_dir = os.path.abspath(
            os.path.normpath(self.agent_conf.ha_confs_path))

        return os.path.join(confs_dir, self.resource_id)

    def get_pid_file_path(self):
        return self.get_full_config_file_path('conntrackd.pid')

    def get_lockfile_path(self):
        return self.get_full_config_file_path('conntrackd.lock')

    def get_ctlfile_path(self):
        return self.get_full_config_file_path(
            'conntrackd.ctl',
            maxlen=UNIX_PATH_MAX,
        )

    def get_conffile_path(self):
        return self.get_full_config_file_path('conntrackd.conf')

    def create_pid_file(self):
        config_path = self.get_conffile_path()
        pid_file = self.get_pid_file_path()

        cmd = 'conntrackd -d -C %s' % config_path
        pid = utils.pgrep(cmd)

        if not pid:
            raise RuntimeError(_('No process for "%s" found.') % cmd)

        file_utils.replace_file(pid_file, pid)

    def spawn(self):
        config_path = self.output_config_file()

        def callback(pidfile):
            cmd = ['conntrackd', '-d',
                   '-C', config_path]
            return cmd

        def pre_cmd_callback():
            # conntrackd.lock and conntrackd.ctl must be removed before
            # starting a new conntrackd.
            utils.delete_if_exists(self.get_lockfile_path(), run_as_root=True)
            utils.delete_if_exists(self.get_ctlfile_path(), run_as_root=True)

        def post_cmd_callback():
            self.create_pid_file()

            # Synchronize connection tracking state with peer
            cmd = ['conntrackd', '-C', config_path, '-n']
            utils.execute(cmd, run_as_root=True, check_exit_code=True)

        pm = self.get_process(callback=callback,
                              pre_cmd_callback=pre_cmd_callback,
                              post_cmd_callback=post_cmd_callback)
        pm.enable(reload_cfg=False)

        self.process_monitor.register(uuid=self.resource_id,
                                      service_name=CONNTRACKD_SERVICE_NAME,
                                      monitored_process=pm)

        LOG.debug('Conntrackd spawned with config %s', config_path)

    def get_process(self, callback=None, pre_cmd_callback=None,
                    post_cmd_callback=None):
        return external_process.ProcessManager(
            cfg.CONF,
            self.resource_id,
            self.namespace,
            default_cmd_callback=callback,
            default_pre_cmd_callback=pre_cmd_callback,
            default_post_cmd_callback=post_cmd_callback,
            pid_file=self.get_pid_file_path())

    def disable(self):
        self.process_monitor.unregister(uuid=self.resource_id,
                                        service_name=CONNTRACKD_SERVICE_NAME)

        pm = self.get_process()
        if not pm.active:
            return

        # First try to stop conntrackd by using its own control command
        config_path = self.get_conffile_path()
        cmd = ['conntrackd', '-C', config_path, '-k']
        utils.execute(cmd, run_as_root=True)

        try:
            common_utils.wait_until_true(lambda: not pm.active,
                                         timeout=SIGTERM_TIMEOUT)
        except common_utils.WaitTimeout:
            LOG.warning('Conntrackd process %s did not finish after asking it '
                        'to shut down in %s seconds, sending SIGKILL signal.',
                        pm.pid, SIGTERM_TIMEOUT)
            pm.disable(sig=str(int(signal.SIGKILL)))

    def build_config(self):
        return CONFIG_TEMPLATE.render(
            hash_size=self.agent_conf.ha_conntrackd_hashsize,
            hash_limit=self.agent_conf.ha_conntrackd_hashlimit,
            lockfile_path=self.get_lockfile_path(),
            socket_path=self.get_ctlfile_path(),
            unix_backlog=self.agent_conf.ha_conntrackd_unix_backlog,
            socket_buffer_size=self.agent_conf.ha_conntrackd_socketbuffersize,
            socket_buffer_size_max_grown=(
                self.agent_conf.ha_conntrackd_socketbuffersize_max_grown
            ),
            protocol_accept=[
                'TCP', 'SCTP', 'DCCP', 'UDP', 'ICMP', 'IPv6-ICMP'
            ],
            # Ignore loopback and HA sync addresses
            address_ignore=[
                (4, '127.0.0.1'),
                (6, '::1'),
                (4, self.mcast_iface_addr),
            ],
            ipv4_mcast_addr=self.agent_conf.ha_conntrackd_ipv4_mcast_addr,
            ipv4_interface=self.mcast_iface_addr,
            mcast_group=self.agent_conf.ha_conntrackd_group + self.ha_vr_id,
            interface=self.ha_iface,
            snd_socket_buffer=self.agent_conf.ha_conntrackd_sndsocketbuffer,
            rcv_socket_buffer=self.agent_conf.ha_conntrackd_rcvsocketbuffer,
        )

    def output_config_file(self):
        config_path = self.get_conffile_path()
        file_utils.replace_file(config_path, self.build_config())

        return config_path



# Copyright (c) 2024
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovs.db import idl
import ovs.ovsuuid


# Temporarily fix ovs.db.idl.Transaction._substitute_uuids to support handling
# the persist_uuid feature
def _substitute_uuids(self, json):
    if isinstance(json, list | tuple):
        if (len(json) == 2 and
                json[0] == 'uuid' and
                ovs.ovsuuid.is_valid_string(json[1])):
            uuid = ovs.ovsuuid.from_string(json[1])
            row = self._txn_rows.get(uuid, None)
            if row and row._data is None and not row._persist_uuid:
                return ["named-uuid", idl._uuid_name_from_uuid(uuid)]
        else:
            return [self._substitute_uuids(elem) for elem in json]
    return json


def apply_ovs_fixes():
    idl.Transaction._substitute_uuids = _substitute_uuids



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron.tests import base

from neutron.privileged.agent.ovsdb.native import helpers


class OvsdbNativeHelpersTestCase(base.BaseTestCase):

    def test__connect_to_manager_uri_ipv4(self):
        self.assertEqual(
            'ptcp:6640:127.0.0.1',
            helpers._connection_to_manager_uri('tcp:127.0.0.1:6640'))
        self.assertEqual(
            'pssl:6640:127.0.0.1',
            helpers._connection_to_manager_uri('ssl:127.0.0.1:6640'))
        self.assertEqual(
            'ptcp:127.0.0.1',
            helpers._connection_to_manager_uri('tcp:127.0.0.1'))

    def test__connect_to_manager_uri_ipv6(self):
        self.assertEqual(
            'ptcp:6640:[::1]',
            helpers._connection_to_manager_uri('tcp:[::1]:6640'))
        self.assertEqual(
            'pssl:6640:[::1]',
            helpers._connection_to_manager_uri('ssl:[::1]:6640'))
        self.assertEqual(
            'ptcp:[::1]',
            helpers._connection_to_manager_uri('tcp:[::1]'))

    def test__connect_to_manager_uri_hostname(self):
        self.assertEqual(
            'ptcp:6640:localhost',
            helpers._connection_to_manager_uri('tcp:localhost:6640'))
        self.assertEqual(
            'pssl:6640:localhost',
            helpers._connection_to_manager_uri('ssl:localhost:6640'))
        self.assertEqual(
            'ptcp:localhost',
            helpers._connection_to_manager_uri('tcp:localhost'))

    def test__connect_to_manager_uri_unix(self):
        self.assertEqual(
            'punix:/run/openvswitch/db.sock',
            helpers._connection_to_manager_uri(
                'unix:/run/openvswitch/db.sock'))



# Copyright (c) 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import secrets
import threading
import uuid

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service
from ovsdbapp.backend.ovs_idl import event as row_event

from neutron.agent.ovn.agent import ovsdb
from neutron.agent.ovn.extensions import extension_manager as ext_mgr
from neutron.common.ovn import constants as ovn_const


LOG = logging.getLogger(__name__)
OVN_MONITOR_UUID_NAMESPACE = uuid.UUID('fd7e0970-7164-11ed-80f0-00000003158a')


class SbGlobalUpdateEvent(row_event.RowEvent):
    """Row update event on SB_Global table.

    This event will trigger the OVN Neutron Agent update of the
    'neutron:ovn-neutron-agent-sb-cfg' key in 'SB_Global', that is used to
    determine the agent status.
    """

    def __init__(self, ovn_agent):
        self.ovn_agent = ovn_agent
        table = 'SB_Global'
        events = (self.ROW_UPDATE, )
        super().__init__(events, table, None)
        self.event_name = self.__class__.__name__
        self._first_run = True

    def run(self, event, row, old):
        def _update_chassis(self, row):
            self.ovn_agent.update_neutron_sb_cfg_key(nb_cfg=row.nb_cfg)

        delay = 0
        if self._first_run:
            self._first_run = False
        else:
            # We occasionally see port binding failed errors due to
            # the ML2 driver refusing to bind the port to a dead agent.
            # If all agents heartbeat at the same time, they will all
            # cause a load spike on the server. To mitigate that it is needed
            # to spread out the load by introducing a random delay.
            max_delay = max(min(cfg.CONF.agent_down_time // 3, 10), 3)
            delay = secrets.SystemRandom().randint(0, max_delay)

        LOG.debug('Delaying updating chassis table for %s seconds', delay)
        timer = threading.Timer(delay, _update_chassis, [self, row])
        timer.start()


class ChassisPrivateCreateEvent(row_event.RowEvent):
    """Row create event - Chassis name == our_chassis.

    On connection, we get a dump of all chassis so if we catch a creation
    of our own chassis it has to be a reconnection. In this case, we need
    to do a full sync to make sure that we capture all changes while the
    connection to OVSDB was down.
    """
    def __init__(self, ovn_agent):
        self._first_time = True
        self.ovn_agent = ovn_agent
        events = (self.ROW_CREATE,)
        super().__init__(events, 'Chassis_Private', None)
        self.conditions = (('name', '=', self.ovn_agent.chassis),)
        self.event_name = self.__class__.__name__

    def run(self, event, row, old):
        if self._first_time:
            self._first_time = False
            return

        # Re-register the OVN agent with the local chassis in case its
        # entry was re-created (happens when restarting the ovn-controller)
        self.ovn_agent.register_ovn_agent()
        self.ovn_agent.update_neutron_sb_cfg_key()


class OVNNeutronAgent(service.Service):

    def __init__(self, conf):
        super().__init__()
        self._conf = conf
        self._chassis = None
        self._chassis_id = None
        self._ovn_bridge = None
        self.ext_manager_api = None
        self.ext_manager = None

    def __getitem__(self, name):
        """Return the named extension objet from ``self.ext_manager``"""
        return self.ext_manager[name].obj

    @property
    def conf(self):
        return self._conf

    @property
    def ovs_idl(self):
        if not self.ext_manager_api.ovs_idl:
            self.ext_manager_api.ovs_idl = self._load_ovs_idl()
        return self.ext_manager_api.ovs_idl

    @property
    def nb_idl(self):
        return self.ext_manager_api.nb_idl

    @property
    def nb_post_fork_event(self):
        return self.ext_manager_api.nb_post_fork_event

    @property
    def sb_idl(self):
        return self.ext_manager_api.sb_idl

    @property
    def sb_post_fork_event(self):
        return self.ext_manager_api.sb_post_fork_event

    @property
    def chassis(self):
        return self._chassis

    @property
    def chassis_id(self):
        return self._chassis_id

    @property
    def ovn_bridge(self):
        return self._ovn_bridge

    def load_config(self):
        self._chassis = ovsdb.get_own_chassis_name(self.ovs_idl)
        try:
            self._chassis_id = uuid.UUID(self.chassis)
        except ValueError:
            # OVS system-id could be a non UUID formatted string.
            self._chassis_id = uuid.uuid5(OVN_MONITOR_UUID_NAMESPACE,
                                          self._chassis)
        self._ovn_bridge = ovsdb.get_ovn_bridge(self.ovs_idl)
        LOG.info("Loaded chassis name %s (UUID: %s) and ovn bridge %s.",
                 self.chassis, self.chassis_id, self.ovn_bridge)

    def _load_ovs_idl(self):
        events = []
        for extension in self.ext_manager:
            events += extension.obj.ovs_idl_events
        events = [e(self) for e in set(events)]
        return ovsdb.MonitorAgentOvsIdl(set(events)).start()

    def _load_nb_idl(self):
        events = []
        tables = []
        for extension in self.ext_manager:
            events += extension.obj.nb_idl_events
            tables += extension.obj.nb_idl_tables

        if not (tables or events):
            # If there is no need to retrieve any table nor attend to any
            # event, the IDL object is not created to save a DB connection.
            return None

        events = [e(self) for e in set(events)]
        tables = set(tables)
        return ovsdb.MonitorAgentOvnNbIdl(tables, events).start()

    def _load_sb_idl(self):
        events = [SbGlobalUpdateEvent,
                  ChassisPrivateCreateEvent,
                  ]
        tables = ['SB_Global', 'Chassis_Private']
        for extension in self.ext_manager:
            events += extension.obj.sb_idl_events
            tables += extension.obj.sb_idl_tables

        events = [e(self) for e in set(events)]
        tables = set(tables)
        return ovsdb.MonitorAgentOvnSbIdl(tables, events,
                                          chassis=self.chassis).start()

    def register_ovn_agent(self):
        # NOTE(lucasagomes): db_add() will not overwrite the UUID if
        # it's already set.
        # Generate unique, but consistent ovn agent id for chassis name
        agent_id = uuid.uuid5(self.chassis_id, 'ovn_agent')
        ext_ids = {ovn_const.OVN_AGENT_NEUTRON_ID_KEY: str(agent_id)}
        self.sb_idl.db_add('Chassis_Private', self.chassis, 'external_ids',
                           ext_ids).execute(check_error=True)

    def _cleanup_previous_tags(self):
        """Remove any existing tag related to the OVN Metadata agent

        The OVN Metadata agent is deprecated and marked for removal in 2026.2.
        This code should stay during the following SLURP release (2027.1) and
        be removed in the next release (2027.2).

        While both agents can provide the same functionality (OVN Metadata
        agent and OVN agent with the metadata extension), it is needed to
        provide a cleanup method for any leftover tag from the other agent.
        """
        metadata_keys = (ovn_const.OVN_AGENT_METADATA_SB_CFG_KEY,
                         ovn_const.OVN_AGENT_METADATA_DESC_KEY,
                         ovn_const.OVN_AGENT_METADATA_ID_KEY)
        self.sb_idl.db_remove(
            'Chassis_Private', self.chassis, 'external_ids',
            *metadata_keys, if_exists=True).execute(check_error=True)

    def update_neutron_sb_cfg_key(self, nb_cfg=None):
        nb_cfg = (nb_cfg or
                  self.sb_idl.db_get('Chassis_Private',
                                     self.chassis, 'nb_cfg').execute())
        external_ids = {ovn_const.OVN_AGENT_NEUTRON_SB_CFG_KEY: str(nb_cfg)}
        self.sb_idl.db_set(
            'Chassis_Private', self.chassis,
            ('external_ids', external_ids)).execute(check_error=True)

    def _initialize_ext_manager(self):
        """Initialize the externsion manager and the extension manager API.

        This method must be called once, outside the ``__init__`` method and
        at the beginning of the ``start`` method.
        """
        if not self.ext_manager:
            self.ext_manager_api = ext_mgr.OVNAgentExtensionAPI()
            self.ext_manager = ext_mgr.OVNAgentExtensionManager(self._conf)
            self.ext_manager.initialize(None, 'ovn', self)

    def start(self):
        # This must be the first operation in the `start` method.
        self._initialize_ext_manager()

        # Extension manager configuration.
        self.ext_manager_api.ovs_idl = self._load_ovs_idl()
        self.load_config()
        # Before executing "_load_sb_idl", is is needed to execute
        # "load_config" to populate self.chassis.
        self.ext_manager_api.sb_idl = self._load_sb_idl()
        self.ext_manager_api.nb_idl = self._load_nb_idl()
        self.ext_manager.start()

        self._cleanup_previous_tags()
        self.register_ovn_agent()
        self.update_neutron_sb_cfg_key()
        LOG.info('OVN Neutron Agent started')

    def stop(self, graceful=True):
        LOG.info('Stopping OVN Neutron Agent')
        super().stop(graceful)



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Initial operations for the Mellanox plugin

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'brocadenetworks',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('vlan', sa.String(length=10), nullable=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'brocadeports',
        sa.Column('port_id', sa.String(length=36), nullable=False,
                  server_default=''),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('admin_state_up', sa.Boolean(), nullable=False),
        sa.Column('physical_interface', sa.String(length=36), nullable=True),
        sa.Column('vlan_id', sa.String(length=36), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['network_id'], ['brocadenetworks.id'], ),
        sa.PrimaryKeyConstraint('port_id'))

    op.create_table(
        'ml2_brocadenetworks',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('vlan', sa.String(length=10), nullable=True),
        sa.Column('segment_id', sa.String(length=36), nullable=True),
        sa.Column('network_type', sa.String(length=10), nullable=True),
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.PrimaryKeyConstraint('id'))

    op.create_table(
        'ml2_brocadeports',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('admin_state_up', sa.Boolean(), nullable=False),
        sa.Column('physical_interface', sa.String(length=36), nullable=True),
        sa.Column('vlan_id', sa.String(length=36), nullable=True),
        sa.Column('tenant_id', sa.String(length=255), nullable=True,
                  index=True),
        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(['network_id'], ['ml2_brocadenetworks.id']))



# Copyright 2024 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json

from alembic import op
from oslo_serialization import jsonutils
from oslo_utils import strutils
import sqlalchemy as sa

from neutron.db import migration


# Add port trusted attribute
#
# Revision ID: 5bcb7b31ec7d
# Revises: 175fa80908e1
# Create Date: 2024-08-06 12:44:37.193211

# revision identifiers, used by Alembic.
revision = '5bcb7b31ec7d'
down_revision = '175fa80908e1'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.RELEASE_2024_2]


def upgrade():
    port_trusted_table = migration.create_table_if_not_exists(
        'porttrusted',
        sa.Column('port_id',
                  sa.String(36),
                  sa.ForeignKey('ports.id',
                                ondelete="CASCADE"),
                  primary_key=True),
        sa.Column('trusted',
                  sa.Boolean,
                  nullable=True))

    if port_trusted_table is None:
        # Table was already created before so no need to insert any data
        # to it now
        return

    # A simple model of the ml2_port_bindings table, just to get and update
    # binding:profile fields where needed
    port_binding_table = sa.Table(
        'ml2_port_bindings', sa.MetaData(),
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('profile', sa.String(length=4095)))

    session = sa.orm.Session(bind=op.get_bind())
    for row in session.query(port_binding_table).all():
        if len(row[1]) == 0:
            continue
        try:
            profile = jsonutils.loads(row[1])
        except json.JSONDecodeError:
            continue
        trusted = profile.pop('trusted', None)
        if trusted is None:
            continue
        session.execute(port_trusted_table.insert().values(
            port_id=row[0],
            trusted=strutils.bool_from_string(trusted)))
        session.execute(port_binding_table.update().values(
            profile=jsonutils.dumps(profile) if profile else '').where(
                port_binding_table.c.port_id == row[0]))
    session.commit()



# Copyright (c) 2024 Red Hat Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import uplink_status_propagation_updatable \
    as uspu
from neutron_lib.plugins.ml2 import api
from oslo_log import log as logging

from neutron.db import uplink_status_propagation_db as usp_db


LOG = logging.getLogger(__name__)


class UplinkStatusPropagationUpdatableExtensionDriver(
        api.ExtensionDriver, usp_db.UplinkStatusPropagationMixin):

    _supported_extension_alias = uspu.ALIAS

    def initialize(self):
        LOG.info('UplinkStatusPropagationUpdatableExtensionDriver '
                 'initialization complete')

    @property
    def extension_alias(self):
        return self._supported_extension_alias

    def process_update_port(self, context, data, result):
        if uspu.PROPAGATE_UPLINK_STATUS in data:
            self._process_update_port(context, data, result)



# Copyright 2017 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.metadata import driver_base

_HEADER_CONFIG_TEMPLATE = """
    http-request add-header X-OVN-%(res_type)s-ID %(res_id)s
"""


class HaproxyConfigurator(driver_base.HaproxyConfiguratorBase):
    PROXY_CONFIG_DIR = "ovn-metadata-proxy"
    HEADER_CONFIG_TEMPLATE = _HEADER_CONFIG_TEMPLATE


class MetadataDriver(driver_base.MetadataDriverBase):
    @staticmethod
    def haproxy_configurator():
        return HaproxyConfigurator



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile
from unittest import mock
import weakref

import netaddr
from oslo_config import cfg
from ovsdbapp import venv

from neutron.agent.ovsdb import impl_idl
from neutron.common import utils
from neutron.services.bgp import constants
from neutron.tests.common.exclusive_resources import ip_address
from neutron.tests.common import net_helpers
from neutron.tests.functional.agent.ovn.agent import test_ovn_neutron_agent
from neutron.tests.functional.agent.ovn.extensions import bgp as test_bgp


class FakeLoopbackContext:
    """Context manager that creates a fake loopback device and mocks constants.

    This creates a veth pair to use as a fake loopback device and patches
    the agent hostdev_name and netaddr.is_loopback function to consider our
    addresses as loopback.
    """

    def __init__(self, test_case):
        """Initialize the fake loopback context.

        :param test_case: The test case instance, used to register fixtures.
        """
        self.test_case = test_case
        self.device = None
        self._patches = []

    def __enter__(self):
        # Create a fake loopback device using a veth pair
        self.device = self.test_case.useFixture(
            net_helpers.VethFixture()).ports[0]
        self.device.link.set_up()

        # Get an exclusive IP for the fake localhost address
        fake_localhost = self.test_case.useFixture(
            ip_address.ExclusiveIPAddress('169.254.0.1', '169.254.0.254'))

        # Add the fake localhost IP to the device
        self.device.addr.add(f'{fake_localhost.address}/8',
                             add_broadcast=False)

        @property
        def is_loopback_mock(ip_network):
            return lambda: str(ip_network) == f"{fake_localhost.address}/8"

        # Start the patches
        self._patches = [
            mock.patch.object(
                self.test_case.bgp_ext, 'hostdev_name', self.device.name),
            mock.patch.object(netaddr.IPNetwork, 'is_loopback',
                              is_loopback_mock),
        ]
        for p in self._patches:
            p.start()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for p in self._patches:
            p.stop()
        return False

    def add_ips(self, cidrs):
        """Add IP addresses to the fake loopback device.

        :param cidrs: List of CIDR strings to add to the device.
        """
        for cidr in cidrs:
            self.device.addr.add(cidr, add_broadcast=False)


class BGPExtensionBaseTestCase(test_ovn_neutron_agent.TestOVNNeutronAgentBase):
    def setUp(self, **kwargs):
        test_ovn_neutron_agent.EXTENSION_NAMES[
            constants.AGENT_BGP_EXT_NAME] = 'BGP agent extension'
        try:
            super().setUp(extensions=[constants.AGENT_BGP_EXT_NAME], **kwargs)
        finally:
            self.addCleanup(self._reset_class_attributes)

    def _reset_class_attributes(self):
        # EXTENSION_NAMES is a class attribute so we need to reset it for other
        # tests running in the same process
        try:
            del test_ovn_neutron_agent.EXTENSION_NAMES[
                constants.AGENT_BGP_EXT_NAME]
        except KeyError:
            pass

    @property
    def bgp_ext(self):
        return self.ovn_agent[constants.AGENT_BGP_EXT_NAME]


class BGPExtensionTestCase(BGPExtensionBaseTestCase):
    def setUp(self, **kwargs):
        self.ovs_venv = self.useFixture(venv.OvsVenvFixture(
            tempfile.mkdtemp(),
            remove=True,
        ))
        _orig_ovsdb_connection = cfg.CONF.OVS.ovsdb_connection
        cfg.CONF.set_override(
            'ovsdb_connection', self.ovs_venv.ovs_connection, group='OVS')
        self.addCleanup(
            cfg.CONF.set_override,
            'ovsdb_connection',
            _orig_ovsdb_connection,
            group='OVS')
        # Cleanup after tests that did not cleanup after themselves
        self._reset_class_attributes()
        super().setUp(**kwargs)

    def _reset_class_attributes(self):
        # The connection is a class attribute so we need to reset it in order
        # to connect to the newly spawned per-test ovsdb server
        impl_idl.NeutronOvsdbIdl._klass._ovsdb_connection = None

        # We spawn a different ovsdb server for each test
        # let's make sure the connection is always a new one
        impl_idl._connection = None

        # We also need to reset the SingletonDecorator
        utils.SingletonDecorator._singleton_instances = (
            weakref.WeakValueDictionary())

        super()._reset_class_attributes()

    def _add_bgp_bridge(self, bridge_name):
        # Get current BGP bridges and append the new one
        ext_ids = self.ovn_agent.ovs_idl.db_get(
            'Open_vSwitch', '.', 'external_ids').execute(check_error=True)
        current_bridges = ext_ids.get(constants.AGENT_BGP_PEER_BRIDGES, '')
        if current_bridges:
            new_bridges = f'{current_bridges},{bridge_name}'
        else:
            new_bridges = bridge_name

        with self.ovn_agent.ovs_idl.transaction(check_error=True) as txn:
            txn.add(self.ovn_agent.ovs_idl.add_br(bridge_name))
            txn.add(self.ovn_agent.ovs_idl.db_set(
                'Open_vSwitch', '.',
                external_ids={constants.AGENT_BGP_PEER_BRIDGES: new_bridges}))
            txn.add(self.ovn_agent.ovs_idl.add_port(
                bridge_name, f'fake-nic-{bridge_name}', type=''))

        # Wait until the agent picks up the bridge name
        utils.wait_until_true(
            lambda: bridge_name in self.bgp_ext.bgp_bridges,
            timeout=10, exception=Exception(
                'Bridge %s not added or not detected by '
                'BGPChassisBridge' % bridge_name))

        bgp_bridge = self.bgp_ext.bgp_bridges[bridge_name]

        return bgp_bridge

    def _get_bridge_mappings(self):
        return self.ovn_agent.ovs_idl.db_get(
            'Open_vSwitch', '.',
            'external_ids').execute(
                check_error=True).get('ovn-bridge-mappings', '')

    def _check_bridge_mappings(self, expected_bms):
        def wait_for_bms():
            bms = self._get_bridge_mappings()
            return sorted(bms.split(',')) == sorted(expected_bms.split(','))

        utils.wait_until_true(
            wait_for_bms,
            sleep=0.1,
            timeout=5,
            exception=Exception(
                "Expected bridge mappings %s were not configured" %
                expected_bms)
        )

    def _verify_chassis_bgp_bridges(self, expected_bridge_names):
        def wait_for_chassis_bgp_bridges():
            ext_ids = self.ovn_agent.sb_idl.db_get(
                'Chassis_Private', self.chassis_name, 'external_ids').execute(
                    check_error=True)
            try:
                observed_bridges = ext_ids[
                    constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY]
            except KeyError:
                observed_bridges = ''
            return observed_bridges == expected_bridge_names

        utils.wait_until_true(
            wait_for_chassis_bgp_bridges,
            sleep=0.1,
            timeout=5,
            exception=Exception(
                "Expected chassis BGP bridges %s were not configured" %
                expected_bridge_names))

    def test_bgp_extension_configures_bridges(self):
        self.ovn_agent.ovs_idl.db_set(
            'Open_vSwitch', '.',
            external_ids={
                'ovn-bridge-mappings': 'physnet:bridge',
                constants.AGENT_BGP_PEER_BRIDGES: 'bgp-br-1,bgp-br-2'}
        ).execute(check_error=True)

        expected_bms = 'physnet:bridge,bgp-br-1:bgp-br-1,bgp-br-2:bgp-br-2'
        self._check_bridge_mappings(expected_bms)
        self._verify_chassis_bgp_bridges('bgp-br-1,bgp-br-2')

    def test_bgp_extension_configures_bridge_with_empty_bms(self):
        self.ovn_agent.ovs_idl.db_set(
            'Open_vSwitch', '.',
            external_ids={
                constants.AGENT_BGP_PEER_BRIDGES: 'bgp-br-1,bgp-br-2'}
            ).execute(check_error=True)

        expected_bms = 'bgp-br-1:bgp-br-1,bgp-br-2:bgp-br-2'
        self._check_bridge_mappings(expected_bms)
        self._verify_chassis_bgp_bridges('bgp-br-1,bgp-br-2')

    def test_bgp_extension_missing_bgp_peer_bridges(self):
        self.ovn_agent.ovs_idl.db_set(
            'Open_vSwitch', '.',
            external_ids={'ovn-bridge-mappings': 'physnet:bridge'}).execute(
                check_error=True)

        expected_bms = 'physnet:bridge'
        self._check_bridge_mappings(expected_bms)
        self._verify_chassis_bgp_bridges('')

    def _test_lrp_with_mac_helper(self, bridge_name):
        port_name = 'ovn-port-bgp'
        lrp_mac = '02:00:00:00:00:00'
        pb_wait_event = test_bgp.WaitForPortBindingEvent(port_name)
        self.ovn_agent.sb_idl.idl.notify_handler.watch_event(pb_wait_event)

        lrp_ext_ids = {constants.LRP_NETWORK_NAME_EXT_ID_KEY: bridge_name}

        bgp_bridge = self._add_bgp_bridge(bridge_name)

        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.lr_add('lr-bgp',
                                  options={'chassis': self.chassis_name}))
            txn.add(self.nb_api.lrp_add(
                'lr-bgp',
                port_name,
                mac=lrp_mac,
                networks=['192.168.1.2/30'],
                external_ids=lrp_ext_ids))

        self.assertTrue(pb_wait_event.wait())

        try:
            pb = self.ovn_agent.sb_idl.db_find_rows(
                'Port_Binding', ('logical_port', '=', port_name)).execute(
                    check_error=True)[0]
        except IndexError:
            self.fail('Port binding for port %s not found' % port_name)

        return (bgp_bridge, pb.mac[0].split(' ', 1)[0])

    def test_new_lrp_with_mac(self):
        bridge_name = 'ovn-br-bgp'
        bgp_bridge, mac = self._test_lrp_with_mac_helper(bridge_name)
        utils.wait_until_true(
            lambda: mac == bgp_bridge.lrp_mac,
            sleep=0.1,
            timeout=5,
            exception=Exception(
                "Expected LRP MAC %s was not configured, is %s" %
                (mac, bgp_bridge.lrp_mac)))

    def test_existing_lrp_with_mac(self):
        bridge_name = 'ovn-br-bgp'
        bgp_bridge, mac = self._test_lrp_with_mac_helper(bridge_name)

        del self.bgp_ext.bgp_bridges[bridge_name]

        bgp_bridge = self.bgp_ext.create_bgp_bridge(bridge_name)

        self.assertEqual(mac, bgp_bridge.lrp_mac)

    def test_watch_patch_port_created_event(self):
        bgp_bridge = self._add_bgp_bridge('ovn-br-bgp')
        self.bgp_ext.watch_port_created_event(bgp_bridge, 'patch')

        self.ovn_agent.ovs_idl.add_port(
            'ovn-br-bgp', 'ovn-port-bgp', type='patch').execute(
                check_error=True)

        utils.wait_until_true(
            lambda: bgp_bridge.patch_port_ofport is not None,
            timeout=5,
            exception=Exception(
                "Patch port ofport was not configured"))

    def test_creating_bgp_bridge_with_existing_patch_port(self):
        bridge_name = 'ovn-br-bgp'
        bgp_bridge = self._add_bgp_bridge(bridge_name)
        self.ovn_agent.ovs_idl.add_port(
            bridge_name, 'ovn-port-bgp', type='patch').execute(
                check_error=True)

        del self.bgp_ext.bgp_bridges[bridge_name]

        bgp_bridge = self.bgp_ext.create_bgp_bridge(bridge_name)

        self.assertIsNotNone(bgp_bridge.patch_port_ofport)


class BGPExtensionHostdevIpsTestCase(BGPExtensionBaseTestCase):
    def _check_host_ips(self, loopback_ips):
        """Helper to test host_ips with given IP configurations.

        Creates a fake loopback device and two BGP bridges with the specified
        IPs, then verifies that host_ips returns all expected IPs.

        :param loopback_ips: List of CIDRs to add to the fake loopback device.
        :param bridge1_ips: List of CIDRs to add to the first bridge.
        :param bridge2_ips: List of CIDRs to add to the second bridge.
        """
        with FakeLoopbackContext(self) as fake_lo:
            fake_lo.add_ips(loopback_ips)

            host_ips = self.bgp_ext.hostdev_ips
            host_ip_cidrs = [str(ip) for ip in host_ips]

            self.assertCountEqual(loopback_ips, host_ip_cidrs)

    def _get_exclusive_ip(self, test_net_number):
        """Get an exclusive IP address from a TEST-NET range.

        :param test_net_number: 1, 2, or 3 for different TEST-NET ranges.
        :returns: CIDR string with /32 prefix.
        """
        exclusive_ip = self.useFixture(
            ip_address.get_test_net_address_fixture(test_net_number))
        return f'{exclusive_ip.address}/32'

    def test_hostdev_ips_combines_loopback(self):
        loopback_ips = [
            self._get_exclusive_ip(1),
            self._get_exclusive_ip(1),
        ]

        self._check_host_ips(loopback_ips)

    def test_hostdev_ips_with_loopback_having_no_ips(self):
        loopback_ips = []

        self._check_host_ips(loopback_ips)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
from unittest import mock

import testtools

from neutron.agent.ovn.extensions.bgp import events
from neutron.common import utils
from neutron.services.bgp import constants
from neutron.tests.common import net_helpers
from neutron.tests.functional.services import bgp as bgp_base


class BridgeNotMatchedException(Exception):
    def __init__(self):
        super().__init__("BGP bridge event was not matched")


class BaseBgpEventsTestCase(bgp_base.BaseBgpIDLTestCase):
    schemas = ['Open_vSwitch']


class FakeAgentAPI:
    def __init__(self, ovs_idl):
        self.ovs_idl = ovs_idl

    def set_chassis_bgp_bridges(self, bridge_names):
        pass


class NewBgpBridgeEventTestCase(BaseBgpEventsTestCase):
    def setUp(self):
        super().setUp()
        self.bgp_ext = mock.Mock()
        self.agent_api = {constants.AGENT_BGP_EXT_NAME: self.bgp_ext}

    def _register_event(self):
        self.ovs_api.idl.notify_handler.watch_event(
            events.NewBgpBridgeEvent(self.agent_api))

    def _set_bgp_bridges(self, bgp_bridges):
        self.ovs_api.db_set(
            'Open_vSwitch', '.',
            external_ids={constants.AGENT_BGP_PEER_BRIDGES: bgp_bridges}
        ).execute(check_error=True)

    def _create_fake_nic(self):
        return self.useFixture(net_helpers.VethFixture()).ports[0]

    def _check_event_not_triggered(self):
        with testtools.ExpectedException(BridgeNotMatchedException):
            utils.wait_until_true(
                lambda: self.bgp_ext.create_bgp_bridge.called,
                sleep=0.5,
                timeout=2,
                exception=BridgeNotMatchedException())

    def test_new_bgp_bridge_event(self):
        self._register_event()
        bgp_bridge_name = 'br-bgp'
        self._set_bgp_bridges(bgp_bridge_name)
        fake_nic = self._create_fake_nic()

        with self.ovs_api.transaction(check_error=True) as txn:
            txn.add(self.ovs_api.add_br(bgp_bridge_name))
            txn.add(self.ovs_api.add_port(bgp_bridge_name, fake_nic.name))
        utils.wait_until_true(
            lambda: self.bgp_ext.create_bgp_bridge.called,
            sleep=0.5,
            timeout=5,
            exception=BridgeNotMatchedException())

    def test_new_bgp_bridge_event_interface_added(self):
        bgp_bridge_name = 'br-bgp'
        self._set_bgp_bridges(bgp_bridge_name)
        self.ovs_api.add_br(bgp_bridge_name).execute(check_error=True)
        # let's make sure the event hasn't been triggered
        self._register_event()

        fake_nic = self._create_fake_nic()
        self.ovs_api.add_port(bgp_bridge_name, fake_nic.name).execute(
            check_error=True)
        utils.wait_until_true(
            lambda: self.bgp_ext.create_bgp_bridge.called,
            sleep=0.5,
            timeout=5,
            exception=Exception("BGP bridge %s not matched" % bgp_bridge_name))

    def test_adding_bridge_without_nic_does_not_trigger_event(self):
        self._register_event()
        bgp_bridge_name = 'br-bgp'
        self._set_bgp_bridges(bgp_bridge_name)
        self.ovs_api.add_br(bgp_bridge_name).execute(check_error=True)
        self._check_event_not_triggered()

    def test_modifying_bridge_does_not_trigger_event(self):
        self._register_event()
        bgp_bridge_name = 'br-bgp'
        self._set_bgp_bridges(bgp_bridge_name)
        self.ovs_api.add_br(bgp_bridge_name).execute(check_error=True)

        self.ovs_api.db_set(
            'Bridge', bgp_bridge_name, other_config={'foo': 'bar'}
        ).execute(check_error=True)
        self._check_event_not_triggered()

    def test_removing_nic_does_not_trigger_event(self):
        bgp_bridge_name = 'br-bgp'
        self._set_bgp_bridges(bgp_bridge_name)
        with self.ovs_api.transaction(check_error=True) as txn:
            txn.add(self.ovs_api.add_br(bgp_bridge_name))
            fake_nic = self._create_fake_nic()
            txn.add(self.ovs_api.add_port(bgp_bridge_name, fake_nic.name))
            fake_nic = self._create_fake_nic()
            txn.add(self.ovs_api.add_port(bgp_bridge_name, fake_nic.name))
        self._register_event()

        self.ovs_api.del_port(fake_nic.name).execute(check_error=True)
        self._check_event_not_triggered()

    def test_new_non_bgp_bridge_does_not_trigger_event(self):
        self._register_event()
        bgp_bridge_name = 'br-bgp'
        fake_nic = self._create_fake_nic()

        with self.ovs_api.transaction(check_error=True) as txn:
            txn.add(self.ovs_api.add_br(bgp_bridge_name))
            txn.add(self.ovs_api.add_port(bgp_bridge_name, fake_nic.name))
        self._check_event_not_triggered()

    def test_adding_patch_port_does_not_trigger_event(self):
        bgp_bridge_name = 'br-bgp'
        patch_port_bridge = 'br-patch'
        self._set_bgp_bridges(bgp_bridge_name)
        fake_nic = self._create_fake_nic()

        with self.ovs_api.transaction(check_error=True) as txn:
            txn.add(self.ovs_api.add_br(bgp_bridge_name))
            txn.add(self.ovs_api.add_br(patch_port_bridge))
            txn.add(self.ovs_api.add_port(bgp_bridge_name, fake_nic.name))

        self._register_event()

        with self.ovs_api.transaction(check_error=True) as txn:
            txn.add(self.ovs_api.add_port(bgp_bridge_name, 'bgp-patch-port'))
            txn.add(self.ovs_api.add_port(
                patch_port_bridge, 'patch-patch-port'))
            txn.add(self.ovs_api.db_set(
                'Interface', 'bgp-patch-port', type='patch',
                options={'peer': 'patch-patch-port'}))
            txn.add(self.ovs_api.db_set(
                'Interface', 'patch-patch-port', type='patch',
                options={'peer': 'bgp-patch-port'}))
        self._check_event_not_triggered()


class BgpBridgeMappingsBase(BaseBgpEventsTestCase):
    def setUp(self):
        super().setUp()
        self.agent_api = FakeAgentAPI(self.ovs_api)

    def _verify_mappings(self, key, expected_mappings):
        def wait_for_mappings():
            try:
                mappings = self.ovs_api.db_get(
                    'Open_vSwitch', '.', 'external_ids'
                ).execute(check_error=True)[key]
            except KeyError:
                mappings = ''
            if mappings:
                mappings = sorted(mappings.split(','))
            return mappings == sorted(expected_mappings)

        utils.wait_until_true(
            wait_for_mappings,
            sleep=0.1,
            timeout=5,
            exception=Exception(
                f"Expected {key} {expected_mappings} were not configured",
            ))

    def _verify_bridge_mappings(self, expected_bridge_mappings):
        self._verify_mappings('ovn-bridge-mappings', expected_bridge_mappings)

    def _verify_port_mappings(self, expected_port_mappings):
        self._verify_mappings(
            constants.OVN_DYNAMIC_ROUTING_PORT_MAPPING, expected_port_mappings)


class CreateLocalOVSEventTestCase(BgpBridgeMappingsBase):
    def setUp(self):
        super().setUp()
        self.ovs_api.idl.notify_handler.watch_event(
            events.CreateLocalOVSEvent(self.agent_api))

    def trigger_event(self):
        self.ovs_api.restart_connection()

    def test_create_local_ovs_event(self):
        self.ovs_api.db_set(
            'Open_vSwitch', '.',
            external_ids={
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-1,br-bgp-2'}
        ).execute(check_error=True)
        self.trigger_event()
        expected_mappings = ['br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2']
        self._verify_bridge_mappings(expected_mappings)
        self._verify_port_mappings(expected_mappings)

    def test_create_local_ovs_event_existing_bridge_mappings(self):
        self.ovs_api.db_set(
            'Open_vSwitch', '.',
            external_ids={
                'ovn-bridge-mappings': 'physnet:bridge',
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-1,br-bgp-2'}
        ).execute(check_error=True)
        self.trigger_event()
        expected_bridge_mappings = [
            'physnet:bridge', 'br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2']
        expected_port_mappings = [
            'br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2']
        self._verify_bridge_mappings(expected_bridge_mappings)
        self._verify_port_mappings(expected_port_mappings)

    def test_create_local_ovs_event_existing_bgp_in_bridge_mappings(self):
        self.ovs_api.db_set(
            'Open_vSwitch', '.',
            external_ids={
                'ovn-bridge-mappings': 'physnet:bridge,br-bgp-1:br-bgp-1',
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-1,br-bgp-2'}
        ).execute(check_error=True)
        self.trigger_event()
        expected_bridge_mappings = [
            'physnet:bridge', 'br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2']
        expected_port_mappings = [
            'br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2']
        self._verify_bridge_mappings(expected_bridge_mappings)
        self._verify_port_mappings(expected_port_mappings)


class UpdateLocalOVSEventTestCase(BgpBridgeMappingsBase):
    def _test_helper(
            self, initial_ext_ids,
            new_ext_ids,
            expected_bridge_mappings,
            expected_port_mappings):
        self.ovs_api.db_set(
            'Open_vSwitch', '.',
            external_ids=initial_ext_ids
        ).execute(check_error=True)
        self.ovs_api.idl.notify_handler.watch_event(
            events.UpdateLocalOVSEvent(self.agent_api))

        self.ovs_api.db_set(
            'Open_vSwitch', '.',
            external_ids=new_ext_ids
        ).execute(check_error=True)

        self._verify_bridge_mappings(expected_bridge_mappings)
        self._verify_port_mappings(expected_port_mappings)

    def test_adding_bgp_bridge(self):
        self._test_helper(
            initial_ext_ids={
                'ovn-bridge-mappings': 'physnet:bridge',
            },
            new_ext_ids={
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-1,br-bgp-2'},
            expected_bridge_mappings=[
                'physnet:bridge', 'br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2'],
            expected_port_mappings=['br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2']
        )

    def test_removing_bgp_bridge(self):
        self._test_helper(
            initial_ext_ids={
                'ovn-bridge-mappings':
                    'physnet:bridge,br-bgp-1:br-bgp-1,br-bgp-2:br-bgp-2',
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-1,br-bgp-2',
            },
            new_ext_ids={constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-2'},
            expected_bridge_mappings=['physnet:bridge', 'br-bgp-2:br-bgp-2'],
            expected_port_mappings=['br-bgp-2:br-bgp-2']
        )

    def test_modifying_bgp_bridge(self):
        self._test_helper(
            initial_ext_ids={
                'ovn-bridge-mappings':
                    'physnet:bridge,br-bgp-1:br-bgp-1,br-bgp-2:br-bgp-2',
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-1,br-bgp-2',
            },
            new_ext_ids={
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-2,br-bgp-3'},
            expected_bridge_mappings=[
                'physnet:bridge', 'br-bgp-2:br-bgp-2', 'br-bgp-3:br-bgp-3'],
            expected_port_mappings=['br-bgp-2:br-bgp-2', 'br-bgp-3:br-bgp-3']
        )

    def test_modifying_bridge_mappings(self):
        self._test_helper(
            initial_ext_ids={
                'ovn-bridge-mappings':
                    'physnet:bridge,br-bgp-1:br-bgp-1,br-bgp-2:br-bgp-2',
                constants.AGENT_BGP_PEER_BRIDGES: 'br-bgp-1,br-bgp-2',
            },
            new_ext_ids={
                'ovn-bridge-mappings': 'physnet:bridge,br-bgp-2:br-bgp-2',
            },
            expected_bridge_mappings=[
                'physnet:bridge', 'br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2'],
            expected_port_mappings=['br-bgp-1:br-bgp-1', 'br-bgp-2:br-bgp-2']
        )

    def test_unrelated_change_does_not_trigger_event(self):
        th_event = threading.Event()

        def event_run(event, row, old):
            th_event.set()

        with mock.patch.object(
                events.UpdateLocalOVSEvent, 'run', side_effect=event_run):
            self.ovs_api.idl.notify_handler.watch_event(
                events.UpdateLocalOVSEvent(self.agent_api))
            self.ovs_api.add_br('br-bgp-test').execute(check_error=True)
            self.assertFalse(th_event.wait(5))



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import socketserver
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as config_fixture
import webob

from neutron.agent.metadata import proxy_base
from neutron.agent.ovn.metadata import server_socket as agent
from neutron.common import metadata as common_metadata
from neutron.tests import base


class ConfFixture(config_fixture.Config):
    pass


class TestMetadataProxyHandler(base.BaseTestCase):
    fake_conf = cfg.CONF
    fake_conf_fixture = ConfFixture(fake_conf)

    def setUp(self):
        super().setUp()
        self.useFixture(self.fake_conf_fixture)
        self.log_p = mock.patch.object(proxy_base, 'LOG')
        self.log = self.log_p.start()
        self.agent_log_p = mock.patch.object(agent, 'LOG')
        self.agent_log = self.agent_log_p.start()
        self.sb_idl = mock.Mock()
        agent.MetadataProxyHandler._conf = self.fake_conf
        agent.MetadataProxyHandler._chassis = 'chassis1'
        agent.MetadataProxyHandler._sb_idl = self.sb_idl
        self.mock_bytesio = mock.patch.object(io, 'BytesIO').start()
        self.mock_fromfile = mock.patch.object(
            webob.Request, 'from_file').start()
        self.mock_sfile = mock.patch.object(
            socketserver, '_SocketWriter').start()

    def test_call(self):
        req = mock.Mock()
        with mock.patch.object(agent.MetadataProxyHandler,
                               '_get_instance_and_project_id') as get_ids, \
                mock.patch.object(agent.MetadataProxyHandler,
                                  '_proxy_request') as proxy:
            get_ids.return_value = ('instance_id', 'project_id')
            proxy.return_value = 'value'
            agent.MetadataProxyHandler(req, 'client_address', 'server')
            self.mock_sfile.return_value.write.assert_called_once_with('value')

    def test_call_no_instance_match(self):
        req = mock.Mock()
        with mock.patch.object(agent.MetadataProxyHandler,
                               '_get_instance_and_project_id') as get_ids, \
                mock.patch.object(agent.MetadataProxyHandler,
                                  '_proxy_request'), \
                mock.patch.object(agent.MetadataProxyHandler,
                                  '_get_instance_id') as get_id:
            get_ids.return_value = None, None
            get_id.return_value = 'net_id', None
            agent.MetadataProxyHandler(req, 'client_address', 'server')
            title = '404 Not Found'
            msg = 'Instance was not found on network net_id.'
            response = common_metadata.encode_http_reponse(title, title, msg)
            self.mock_sfile.return_value.write.assert_called_once_with(
                response)

    def test_call_internal_server_error(self):
        req = mock.Mock()
        with mock.patch.object(agent.MetadataProxyHandler,
                               '_get_instance_and_project_id') as get_ids:
            get_ids.side_effect = RuntimeError
            self.assertRaises(RuntimeError,
                              agent.MetadataProxyHandler,
                              req, 'client_address', 'server')

    def test__get_instance_id_network_id_ipv4(self):
        req = mock.Mock(headers={'X-Forwarded-For': '192.168.1.1',
                                 'X-OVN-Network-ID': 'net_id'})
        self.mock_fromfile.return_value = req
        with mock.patch.object(agent.MetadataProxyHandler,
                               'get_port') as get_port, \
                mock.patch.object(agent.MetadataProxyHandler,
                                  '_proxy_request'):
            get_port.return_value = ('device_id', 'project_id')
            agent.MetadataProxyHandler(req, 'client_address', 'server')
            get_port.assert_called_once_with('192.168.1.1',
                                             network_id='net_id',
                                             remote_mac=None,
                                             router_id=None,
                                             skip_cache=False)

    def test__get_instance_id_network_id_ipv6(self):
        req = mock.Mock(headers={'X-Forwarded-For': '2001:db8::1',
                                 'X-OVN-Network-ID': 'net_id'})
        self.mock_fromfile.return_value = req
        with mock.patch.object(agent.MetadataProxyHandler,
                               'get_port') as get_port, \
                mock.patch.object(agent.MetadataProxyHandler,
                                  '_proxy_request'):
            get_port.return_value = ('device_id', 'project_id')
            agent.MetadataProxyHandler(req, 'client_address', 'server')
            get_port.assert_called_once_with('2001:db8::1',
                                             network_id='net_id',
                                             remote_mac=None,
                                             router_id=None,
                                             skip_cache=False)

    def test__get_instance_id_network_id_ipv6_ll(self):
        req = mock.Mock(headers={'X-Forwarded-For': 'fe80::99',
                                 'X-OVN-Network-ID': 'net_id'})
        self.mock_fromfile.return_value = req
        with mock.patch.object(agent.MetadataProxyHandler,
                               'get_port') as get_port, \
                mock.patch.object(agent.MetadataProxyHandler,
                                  '_proxy_request'):
            get_port.return_value = ('device_id', 'project_id')
            agent.MetadataProxyHandler(req, 'client_address', 'server')
            get_port.assert_called_once_with('fe80::99',
                                             network_id='net_id',
                                             remote_mac='02:00:00:00:00:99',
                                             router_id=None,
                                             skip_cache=False)



# Copyright 2025 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from concurrent import futures

from neutron_lib import constants
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib.objects import exceptions as o_exc
from oslo_config import cfg

from neutron.conf import common as common_config
from neutron.conf.plugins.ml2 import config as ml2_config
from neutron.conf.plugins.ml2.drivers import driver_type as driver_type_config
from neutron.objects import network_segment_range as range_obj
from neutron.plugins.ml2.drivers import type_vlan
from neutron.tests.unit import testlib_api


def _initialize_network_segment_range_support(type_driver, start_time):
    # This method is similar to
    # ``VlanTypeDriverBase.initialize_network_segment_range_support``.
    # The method first deletes the existing default network ranges and then
    # creates the new ones. It also adds an extra second before closing the
    # DB transaction.
    admin_context = context.get_admin_context()
    try:
        with db_api.CONTEXT_WRITER.using(admin_context):
            type_driver._delete_expired_default_network_segment_ranges(
                admin_context, start_time)
            type_driver._populate_new_default_network_segment_ranges(
                admin_context, start_time)
    except o_exc.NeutronDbObjectDuplicateEntry:
        pass


class VlanTypeDriverBaseTestCase(testlib_api.MySQLTestCaseMixin,
                                   testlib_api.SqlTestCase):
    def setUp(self):
        super().setUp()
        cfg.CONF.register_opts(common_config.core_opts)
        ml2_config.register_ml2_plugin_opts()
        driver_type_config.register_ml2_drivers_vlan_opts()
        ml2_config.cfg.CONF.set_override(
            'service_plugins', 'network_segment_range')
        self.min = 1001
        self.max = 1020
        self.net_type = constants.TYPE_VLAN
        self.ranges = [f'phys1:{self.min}:{self.max}',
                       f'phys2:{self.min}:{self.max}',
                       f'phys3:{self.min}:{self.max}',
                       ]
        ml2_config.cfg.CONF.set_override(
            'network_vlan_ranges', self.ranges, group='ml2_type_vlan')
        self.admin_ctx = context.get_admin_context()
        self.type_driver = type_vlan.VlanTypeDriver()
        self.type_driver.initialize()

    def _check_sranges(self, sranges):
        self.assertEqual(len(self.ranges), len(sranges))
        for _srange in sranges:
            self.assertEqual(self.net_type, _srange.network_type)
            self.assertEqual(self.min, _srange.minimum)
            self.assertEqual(self.max, _srange.maximum)
            self.assertIn(_srange.physical_network,
                          ('phys1', 'phys2', 'phys3'))

        self.assertEqual({'phys1': [(self.min, self.max)],
                          'phys2': [(self.min, self.max)],
                          'phys3': [(self.min, self.max)]},
                         self.type_driver._network_vlan_ranges)

    def test_initialize_network_segment_range_support(self):
        # Execute the initialization several times with different start times.
        for start_time in range(3):
            self.type_driver.initialize_network_segment_range_support(
                start_time)
            sranges = range_obj.NetworkSegmentRange.get_objects(self.admin_ctx)
            self._check_sranges(sranges)

    def _test_initialize_nsrange(self, same_init_time=True):
        max_workers = 3
        _futures = []
        with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            if same_init_time:
                # All workers are started at the same init time.
                _futures.append(executor.submit(
                    _initialize_network_segment_range_support,
                    self.type_driver, 0))
            else:
                # All workers have different init times.
                for idx in range(max_workers):
                    _futures.append(executor.submit(
                        _initialize_network_segment_range_support,
                        self.type_driver, idx))
            for _future in _futures:
                _future.result()

        sranges = range_obj.NetworkSegmentRange.get_objects(self.admin_ctx)
        self._check_sranges(sranges)

    def test__initialize_nsrange_support_parallel_exec_same_init_time(self):
        self._test_initialize_nsrange(same_init_time=True)

    def test_initialize_nsrange_support_parallel_exec_diff_init_time(self):
        self._test_initialize_nsrange(same_init_time=False)
